@@ -1,86 +1,18 @@
+use super::{Class, Header, Length, Tag};
 use crate::ber::bitstring_to_u64;
 use crate::ber::integer::*;
 use crate::error::BerError;
 use crate::oid::Oid;
-use nom::bitvec::{order::Msb0, slice::BitSlice};
-use rusticata_macros::newtype_enum;
-use std::convert::AsRef;
-use std::convert::From;
-use std::convert::TryFrom;
-use std::fmt;
-use std::ops::Index;
-use std::vec::Vec;
-
-#[derive(Debug, Copy, Clone, PartialEq, Eq)]
-pub struct BerClassFromIntError(pub(crate) ());
-
-#[derive(Debug, Copy, Clone, PartialEq, Eq)]
-pub struct BerSizeError(pub(crate) ());
-
-/// BER Object class of tag
-#[derive(Debug, Copy, Clone, Eq, PartialEq)]
-#[repr(u8)]
-pub enum BerClass {
-    Universal = 0b00,
-    Application = 0b01,
-    ContextSpecific = 0b10,
-    Private = 0b11,
-}
-
-/// Ber Object Length
-#[derive(Debug, Copy, Clone, Eq, PartialEq)]
-pub enum BerSize {
-    /// Definite form (X.690 8.1.3.3)
-    Definite(usize),
-    /// Indefinite form (X.690 8.1.3.6)
-    Indefinite,
-}
-
-/// BER/DER Tag as defined in X.680 section 8.4
-///
-/// X.690 doesn't specify the maximum tag size so we're assuming that people
-/// aren't going to need anything more than a u32.
-#[derive(Clone, Copy, PartialEq, Eq)]
-pub struct BerTag(pub u32);
-
-newtype_enum! {
-impl debug BerTag {
-    EndOfContent = 0x0,
-    Boolean = 0x1,
-    Integer = 0x2,
-    BitString = 0x3,
-    OctetString = 0x4,
-    Null = 0x05,
-    Oid = 0x06,
-    ObjDescriptor = 0x07,
-    External = 0x08,
-    RealType = 0x09,
-    Enumerated = 0xa,
-    EmbeddedPdv = 0xb,
-    Utf8String = 0xc,
-    RelativeOid = 0xd,
-
-    Sequence = 0x10,
-    Set = 0x11,
-    NumericString = 0x12,
-    PrintableString = 0x13,
-    T61String = 0x14,
-    VideotexString = 0x15,
-
-    Ia5String = 0x16,
-    UtcTime = 0x17,
-    GeneralizedTime = 0x18,
-
-    GraphicString = 25, // 0x19
-    VisibleString = 26, // 0x1a
-    GeneralString = 27, // 0x1b
-
-    UniversalString = 0x1c,
-    BmpString = 0x1e,
-
-    Invalid = 0xff,
-}
-}
+use alloc::borrow::ToOwned;
+use alloc::boxed::Box;
+use alloc::vec::Vec;
+use asn1_rs::ASN1DateTime;
+use asn1_rs::Any;
+#[cfg(feature = "bitvec")]
+use bitvec::{order::Msb0, slice::BitSlice};
+use core::convert::AsRef;
+use core::convert::From;
+use core::ops::Index;
 
 /// Representation of a BER-encoded (X.690) object
 ///
@@ -91,29 +23,10 @@ impl debug BerTag {
 /// tagged values).
 #[derive(Debug, Clone, PartialEq)]
 pub struct BerObject<'a> {
-    pub header: BerObjectHeader<'a>,
+    pub header: Header<'a>,
     pub content: BerObjectContent<'a>,
 }
 
-/// BER object header (identifier and length)
-#[derive(Clone, Debug)]
-pub struct BerObjectHeader<'a> {
-    /// Object class: universal, application, context-specific, or private
-    pub class: BerClass,
-    /// Constructed attribute: 1 if constructed, else 0
-    pub structured: u8,
-    /// Tag number
-    pub tag: BerTag,
-    /// Object length: definite or indefinite
-    pub len: BerSize,
-
-    /// Optionally, the raw encoding of the tag
-    ///
-    /// This is useful in some cases, where different representations of the same
-    /// BER tags have different meanings (BER only)
-    pub raw_tag: Option<&'a [u8]>,
-}
-
 /// BER object content
 #[derive(Debug, Clone, PartialEq)]
 #[allow(clippy::upper_case_acronyms)]
@@ -151,13 +64,13 @@ pub enum BerObjectContent<'a> {
     IA5String(&'a str),
     /// UTF8String: decoded string
     UTF8String(&'a str),
-    /// T61String: raw object bytes
-    T61String(&'a [u8]),
-    /// VideotexString: raw object bytes
-    VideotexString(&'a [u8]),
+    /// T61String: decoded string
+    T61String(&'a str),
+    /// VideotexString: decoded string
+    VideotexString(&'a str),
 
-    /// BmpString: raw object bytes
-    BmpString(&'a [u8]),
+    /// BmpString: decoded string
+    BmpString(&'a str),
     /// UniversalString: raw object bytes
     UniversalString(&'a [u8]),
 
@@ -167,169 +80,24 @@ pub enum BerObjectContent<'a> {
     Set(Vec<BerObject<'a>>),
 
     /// UTCTime: decoded string
-    UTCTime(&'a str),
+    UTCTime(ASN1DateTime),
     /// GeneralizedTime: decoded string
-    GeneralizedTime(&'a str),
+    GeneralizedTime(ASN1DateTime),
 
-    /// Object descriptor: raw object bytes
-    ObjectDescriptor(&'a [u8]),
-    /// GraphicString: raw object bytes
-    GraphicString(&'a [u8]),
-    /// GeneralString: raw object bytes
-    GeneralString(&'a [u8]),
+    /// Object descriptor: decoded string
+    ObjectDescriptor(&'a str),
+    /// GraphicString: decoded string
+    GraphicString(&'a str),
+    /// GeneralString: decoded string
+    GeneralString(&'a str),
 
     /// Optional object
     Optional(Option<Box<BerObject<'a>>>),
     /// Tagged object (EXPLICIT): class, tag  and content of inner object
-    Tagged(BerClass, BerTag, Box<BerObject<'a>>),
+    Tagged(Class, Tag, Box<BerObject<'a>>),
 
-    /// Unknown object: object tag (copied from header), and raw content
-    Unknown(BerTag, &'a [u8]),
-}
-
-impl fmt::Display for BerClass {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        let s = match self {
-            BerClass::Universal => "UNIVERSAL",
-            BerClass::Application => "APPLICATION",
-            BerClass::ContextSpecific => "CONTEXT-SPECIFIC",
-            BerClass::Private => "PRIVATE",
-        };
-        write!(f, "{}", s)
-    }
-}
-
-impl From<u32> for BerTag {
-    fn from(v: u32) -> Self {
-        BerTag(v)
-    }
-}
-
-impl BerSize {
-    /// Return true if length is definite and equal to 0
-    pub fn is_null(&self) -> bool {
-        *self == BerSize::Definite(0)
-    }
-
-    /// Get length of primitive object
-    #[inline]
-    pub fn primitive(&self) -> Result<usize, BerError> {
-        match self {
-            BerSize::Definite(sz) => Ok(*sz),
-            BerSize::Indefinite => Err(BerError::IndefiniteLengthUnexpected),
-        }
-    }
-}
-
-impl From<usize> for BerSize {
-    fn from(v: usize) -> Self {
-        BerSize::Definite(v)
-    }
-}
-
-impl TryFrom<u64> for BerSize {
-    type Error = BerSizeError;
-
-    fn try_from(value: u64) -> Result<Self, Self::Error> {
-        let v = usize::try_from(value).or(Err(BerSizeError(())))?;
-        Ok(BerSize::Definite(v))
-    }
-}
-
-impl TryFrom<BerSize> for usize {
-    type Error = BerSizeError;
-
-    #[inline]
-    fn try_from(value: BerSize) -> Result<Self, Self::Error> {
-        match value {
-            BerSize::Definite(sz) => Ok(sz),
-            BerSize::Indefinite => Err(BerSizeError(())),
-        }
-    }
-}
-
-impl TryFrom<u8> for BerClass {
-    type Error = BerClassFromIntError;
-
-    #[inline]
-    fn try_from(value: u8) -> Result<Self, Self::Error> {
-        match value {
-            0b00 => Ok(BerClass::Universal),
-            0b01 => Ok(BerClass::Application),
-            0b10 => Ok(BerClass::ContextSpecific),
-            0b11 => Ok(BerClass::Private),
-            _ => Err(BerClassFromIntError(())),
-        }
-    }
-}
-
-impl<'a> BerObjectHeader<'a> {
-    /// Build a new BER header
-    pub fn new<Len: Into<BerSize>>(class: BerClass, structured: u8, tag: BerTag, len: Len) -> Self {
-        BerObjectHeader {
-            tag,
-            structured,
-            class,
-            len: len.into(),
-            raw_tag: None,
-        }
-    }
-
-    /// Update header class
-    #[inline]
-    pub fn with_class(self, class: BerClass) -> Self {
-        BerObjectHeader { class, ..self }
-    }
-
-    /// Update header tag
-    #[inline]
-    pub fn with_tag(self, tag: BerTag) -> Self {
-        BerObjectHeader { tag, ..self }
-    }
-
-    /// Update header length
-    #[inline]
-    pub fn with_len(self, len: BerSize) -> Self {
-        BerObjectHeader { len, ..self }
-    }
-
-    /// Update header to add reference to raw tag
-    #[inline]
-    pub fn with_raw_tag(self, raw_tag: Option<&'a [u8]>) -> Self {
-        BerObjectHeader { raw_tag, ..self }
-    }
-
-    /// Test if object class is Universal
-    #[inline]
-    pub fn is_universal(&self) -> bool {
-        self.class == BerClass::Universal
-    }
-    /// Test if object class is Application
-    #[inline]
-    pub fn is_application(&self) -> bool {
-        self.class == BerClass::Application
-    }
-    /// Test if object class is Context-specific
-    #[inline]
-    pub fn is_contextspecific(&self) -> bool {
-        self.class == BerClass::ContextSpecific
-    }
-    /// Test if object class is Private
-    #[inline]
-    pub fn is_private(&self) -> bool {
-        self.class == BerClass::Private
-    }
-
-    /// Test if object is primitive
-    #[inline]
-    pub fn is_primitive(&self) -> bool {
-        self.structured == 0
-    }
-    /// Test if object is constructed
-    #[inline]
-    pub fn is_constructed(&self) -> bool {
-        self.structured == 1
-    }
+    /// Private or Unknown (for ex. unknown tag) object
+    Unknown(Any<'a>),
 }
 
 impl<'a> BerObject<'a> {
@@ -337,34 +105,27 @@ impl<'a> BerObject<'a> {
     ///
     /// Note: values are not checked, so the tag can be different from the real content, or flags
     /// can be invalid.
-    pub fn from_header_and_content<'o>(
-        header: BerObjectHeader<'o>,
+    #[inline]
+    pub const fn from_header_and_content<'o>(
+        header: Header<'o>,
         content: BerObjectContent<'o>,
     ) -> BerObject<'o> {
         BerObject { header, content }
     }
 
     /// Build a BerObject from its content, using default flags (no class, correct tag,
-    /// and structured flag set only for Set and Sequence)
-    pub fn from_obj(c: BerObjectContent) -> BerObject {
-        let class = BerClass::Universal;
+    /// and constructed flag set only for Set and Sequence)
+    pub const fn from_obj(c: BerObjectContent) -> BerObject {
+        let class = Class::Universal;
         let tag = c.tag();
-        let structured = match tag {
-            BerTag::Sequence | BerTag::Set => 1,
-            _ => 0,
-        };
-        let header = BerObjectHeader::new(class, structured, tag, BerSize::Definite(0));
+        let constructed = matches!(tag, Tag::Sequence | Tag::Set);
+        let header = Header::new(class, constructed, tag, Length::Definite(0));
         BerObject { header, content: c }
     }
 
     /// Build a DER integer object from a slice containing an encoded integer
-    pub fn from_int_slice(i: &'a [u8]) -> BerObject<'a> {
-        let header = BerObjectHeader::new(
-            BerClass::Universal,
-            0,
-            BerTag::Integer,
-            BerSize::Definite(0),
-        );
+    pub const fn from_int_slice(i: &'a [u8]) -> BerObject<'a> {
+        let header = Header::new(Class::Universal, false, Tag::Integer, Length::Definite(0));
         BerObject {
             header,
             content: BerObjectContent::Integer(i),
@@ -373,32 +134,20 @@ impl<'a> BerObject<'a> {
 
     /// Set a tag for the BER object
     pub fn set_raw_tag(self, raw_tag: Option<&'a [u8]>) -> BerObject {
-        let header = BerObjectHeader {
-            raw_tag,
-            ..self.header
-        };
+        let header = self.header.with_raw_tag(raw_tag.map(|x| x.into()));
         BerObject { header, ..self }
     }
 
     /// Build a DER sequence object from a vector of DER objects
-    pub fn from_seq(l: Vec<BerObject>) -> BerObject {
+    pub const fn from_seq(l: Vec<BerObject>) -> BerObject {
         BerObject::from_obj(BerObjectContent::Sequence(l))
     }
 
     /// Build a DER set object from a vector of DER objects
-    pub fn from_set(l: Vec<BerObject>) -> BerObject {
+    pub const fn from_set(l: Vec<BerObject>) -> BerObject {
         BerObject::from_obj(BerObjectContent::Set(l))
     }
 
-    /// Build a BER header from this object content
-    #[deprecated(
-        since = "0.5.0",
-        note = "please use `obj.header` or `obj.header.clone()` instead"
-    )]
-    pub fn to_header(&self) -> BerObjectHeader {
-        self.header.clone()
-    }
-
     /// Attempt to read a signed integer value from DER object.
     ///
     /// This can fail if the object is not an integer, or if it is too large.
@@ -492,13 +241,13 @@ impl<'a> BerObject<'a> {
 
     /// Attempt to get a reference on the content from an optional object.
     /// This can fail if the object is not optional.
-    pub fn as_optional(&'a self) -> Result<Option<&'_ BerObject<'a>>, BerError> {
+    pub fn as_optional(&self) -> Result<Option<&BerObject<'a>>, BerError> {
         self.content.as_optional()
     }
 
     /// Attempt to get a reference on the content from a tagged object.
     /// This can fail if the object is not tagged.
-    pub fn as_tagged(&'a self) -> Result<(BerClass, BerTag, &'_ BerObject<'a>), BerError> {
+    pub fn as_tagged(&self) -> Result<(Class, Tag, &BerObject<'a>), BerError> {
         self.content.as_tagged()
     }
 
@@ -513,11 +262,12 @@ impl<'a> BerObject<'a> {
 
     /// Attempt to read a BitString value from DER object.
     /// This can fail if the object is not an BitString.
-    pub fn as_bitstring(&'a self) -> Result<BitStringObject<'a>, BerError> {
+    pub fn as_bitstring(&self) -> Result<BitStringObject<'a>, BerError> {
         self.content.as_bitstring()
     }
 
     /// Constructs a shared `&BitSlice` reference over the object data, if available as slice.
+    #[cfg(feature = "bitvec")]
     pub fn as_bitslice(&self) -> Result<&BitSlice<Msb0, u8>, BerError> {
         self.content.as_bitslice()
     }
@@ -545,37 +295,84 @@ impl<'a> BerObject<'a> {
     /// Attempt to get the content from a DER object, as a str.
     /// This can fail if the object does not contain a string type.
     ///
-    /// Only NumericString, VisibleString, UTCTime, GeneralizedTime,
-    /// PrintableString, UTF8String and IA5String are considered here. Other
+    /// Only some string types are considered here. Other
     /// string types can be read using `as_slice`.
     pub fn as_str(&self) -> Result<&'a str, BerError> {
         self.content.as_str()
     }
 
+    /// Get the BER object header's class.
+    #[inline]
+    pub const fn class(&self) -> Class {
+        self.header.class()
+    }
+
+    /// Get the BER object header's tag.
+    #[inline]
+    pub const fn tag(&self) -> Tag {
+        self.header.tag()
+    }
+
+    /// Get the BER object header's length.
+    #[inline]
+    pub const fn length(&self) -> Length {
+        self.header.length()
+    }
+
     /// Test if object class is Universal
-    pub fn is_universal(&self) -> bool {
-        self.header.class == BerClass::Universal
+    #[inline]
+    pub const fn is_universal(&self) -> bool {
+        self.header.is_universal()
     }
     /// Test if object class is Application
-    pub fn is_application(&self) -> bool {
-        self.header.class == BerClass::Application
+    #[inline]
+    pub const fn is_application(&self) -> bool {
+        self.header.is_application()
     }
     /// Test if object class is Context-specific
-    pub fn is_contextspecific(&self) -> bool {
-        self.header.class == BerClass::ContextSpecific
+    #[inline]
+    pub const fn is_contextspecific(&self) -> bool {
+        self.header.is_contextspecific()
     }
     /// Test if object class is Private
+    #[inline]
     pub fn is_private(&self) -> bool {
-        self.header.class == BerClass::Private
+        self.header.is_private()
     }
 
     /// Test if object is primitive
-    pub fn is_primitive(&self) -> bool {
-        self.header.structured == 0
+    #[inline]
+    pub const fn is_primitive(&self) -> bool {
+        self.header.is_primitive()
     }
     /// Test if object is constructed
-    pub fn is_constructed(&self) -> bool {
-        self.header.structured == 1
+    #[inline]
+    pub const fn is_constructed(&self) -> bool {
+        self.header.is_constructed()
+    }
+
+    /// Return error if `class` is not the expected class
+    #[inline]
+    pub const fn assert_class(&self, class: Class) -> Result<(), BerError> {
+        self.header.assert_class(class)
+    }
+
+    /// Return error if `tag` is not the expected tag
+    #[inline]
+    pub const fn assert_tag(&self, tag: Tag) -> Result<(), BerError> {
+        self.header.assert_tag(tag)
+    }
+
+    /// Return error if object is not constructed
+    #[inline]
+    pub const fn assert_constructed(&self) -> Result<(), BerError> {
+        self.header.assert_constructed()
+    }
+
+    /// Return error if object is not primitive
+    #[inline]
+    pub const fn assert_primitive(&self) -> Result<(), BerError> {
+        self.header.assert_primitive()
     }
 }
 
@@ -593,40 +390,6 @@ impl<'a> From<BerObjectContent<'a>> for BerObject<'a> {
     }
 }
 
-/// Replacement function for Option.xor (>= 1.37)
-#[inline]
-pub(crate) fn xor_option<T>(opta: Option<T>, optb: Option<T>) -> Option<T> {
-    match (opta, optb) {
-        (Some(a), None) => Some(a),
-        (None, Some(b)) => Some(b),
-        _ => None,
-    }
-}
-
-/// Compare two BER headers. `len` fields are compared only if both objects have it set (same for `raw_tag`)
-impl<'a> PartialEq<BerObjectHeader<'a>> for BerObjectHeader<'a> {
-    fn eq(&self, other: &BerObjectHeader) -> bool {
-        self.class == other.class
-            && self.tag == other.tag
-            && self.structured == other.structured
-            && {
-                if self.len.is_null() && other.len.is_null() {
-                    self.len == other.len
-                } else {
-                    true
-                }
-            }
-            && {
-                // it tag is present for both, compare it
-                if xor_option(self.raw_tag, other.raw_tag).is_none() {
-                    self.raw_tag == other.raw_tag
-                } else {
-                    true
-                }
-            }
-    }
-}
-
 impl<'a> BerObjectContent<'a> {
     /// Attempt to read a signed integer value from this object.
     ///
@@ -651,7 +414,7 @@ impl<'a> BerObjectContent<'a> {
             };
             Ok(result)
         } else {
-            Err(BerError::InvalidTag)
+            Err(BerError::BerValueError)
         }
     }
 
@@ -678,7 +441,7 @@ impl<'a> BerObjectContent<'a> {
             };
             Ok(result)
         } else {
-            Err(BerError::InvalidTag)
+            Err(BerError::BerValueError)
         }
     }
 
@@ -733,7 +496,7 @@ impl<'a> BerObjectContent<'a> {
             }
             BerObjectContent::BitString(ignored_bits, data) => {
                 bitstring_to_u64(*ignored_bits as usize, data).and_then(|x| {
-                    if x > u64::from(std::u32::MAX) {
+                    if x > u64::from(core::u32::MAX) {
                         Err(BerError::IntegerTooLarge)
                     } else {
                         Ok(x as u32)
@@ -741,7 +504,7 @@ impl<'a> BerObjectContent<'a> {
                 })
             }
             BerObjectContent::Enum(i) => {
-                if *i > u64::from(std::u32::MAX) {
+                if *i > u64::from(core::u32::MAX) {
                     Err(BerError::IntegerTooLarge)
                 } else {
                     Ok(*i as u32)
@@ -770,7 +533,7 @@ impl<'a> BerObjectContent<'a> {
         self.as_oid().map(|o| o.clone())
     }
 
-    pub fn as_optional(&'a self) -> Result<Option<&'_ BerObject<'a>>, BerError> {
+    pub fn as_optional(&self) -> Result<Option<&BerObject<'a>>, BerError> {
         match *self {
             BerObjectContent::Optional(Some(ref o)) => Ok(Some(o)),
             BerObjectContent::Optional(None) => Ok(None),
@@ -778,7 +541,7 @@ impl<'a> BerObjectContent<'a> {
         }
     }
 
-    pub fn as_tagged(&'a self) -> Result<(BerClass, BerTag, &'_ BerObject<'a>), BerError> {
+    pub fn as_tagged(&self) -> Result<(Class, Tag, &BerObject<'a>), BerError> {
         match *self {
             BerObjectContent::Tagged(class, tag, ref o) => Ok((class, tag, o.as_ref())),
             _ => Err(BerError::BerTypeError),
@@ -792,7 +555,7 @@ impl<'a> BerObjectContent<'a> {
         }
     }
 
-    pub fn as_bitstring(&'a self) -> Result<BitStringObject<'a>, BerError> {
+    pub fn as_bitstring(&self) -> Result<BitStringObject<'a>, BerError> {
         match *self {
             BerObjectContent::BitString(_, ref b) => Ok(b.to_owned()),
             _ => Err(BerError::BerTypeError),
@@ -800,9 +563,10 @@ impl<'a> BerObjectContent<'a> {
     }
 
     /// Constructs a shared `&BitSlice` reference over the object data, if available as slice.
+    #[cfg(feature = "bitvec")]
     pub fn as_bitslice(&self) -> Result<&BitSlice<Msb0, u8>, BerError> {
         self.as_slice()
-            .and_then(|s| BitSlice::<Msb0, _>::from_slice(s).ok_or(BerError::BerValueError))
+            .and_then(|s| BitSlice::<Msb0, _>::from_slice(s).map_err(|_| BerError::BerValueError))
     }
 
     pub fn as_sequence(&self) -> Result<&Vec<BerObject<'a>>, BerError> {
@@ -823,23 +587,21 @@ impl<'a> BerObjectContent<'a> {
     pub fn as_slice(&self) -> Result<&'a [u8],BerError> {
         match *self {
             BerObjectContent::NumericString(s) |
-            BerObjectContent::GeneralizedTime(s) |
-            BerObjectContent::UTCTime(s) |
+            BerObjectContent::BmpString(s) |
             BerObjectContent::VisibleString(s) |
             BerObjectContent::PrintableString(s) |
+            BerObjectContent::GeneralString(s) |
+            BerObjectContent::ObjectDescriptor(s) |
+            BerObjectContent::GraphicString(s) |
+            BerObjectContent::T61String(s) |
+            BerObjectContent::VideotexString(s) |
             BerObjectContent::UTF8String(s) |
             BerObjectContent::IA5String(s) => Ok(s.as_ref()),
             BerObjectContent::Integer(s) |
             BerObjectContent::BitString(_,BitStringObject{data:s}) |
             BerObjectContent::OctetString(s) |
-            BerObjectContent::T61String(s) |
-            BerObjectContent::VideotexString(s) |
-            BerObjectContent::BmpString(s) |
-            BerObjectContent::UniversalString(s) |
-            BerObjectContent::ObjectDescriptor(s) |
-            BerObjectContent::GraphicString(s) |
-            BerObjectContent::GeneralString(s) |
-            BerObjectContent::Unknown(_,s) => Ok(s),
+            BerObjectContent::UniversalString(s) => Ok(s),
+            BerObjectContent::Unknown(ref any) => Ok(any.data),
             _ => Err(BerError::BerTypeError),
         }
     }
@@ -848,10 +610,14 @@ impl<'a> BerObjectContent<'a> {
     pub fn as_str(&self) -> Result<&'a str,BerError> {
         match *self {
             BerObjectContent::NumericString(s) |
-            BerObjectContent::GeneralizedTime(s) |
-            BerObjectContent::UTCTime(s) |
+            BerObjectContent::BmpString(s) |
             BerObjectContent::VisibleString(s) |
             BerObjectContent::PrintableString(s) |
+            BerObjectContent::GeneralString(s) |
+            BerObjectContent::ObjectDescriptor(s) |
+            BerObjectContent::GraphicString(s) |
+            BerObjectContent::T61String(s) |
+            BerObjectContent::VideotexString(s) |
             BerObjectContent::UTF8String(s) |
             BerObjectContent::IA5String(s) => Ok(s),
             _ => Err(BerError::BerTypeError),
@@ -859,37 +625,37 @@ impl<'a> BerObjectContent<'a> {
     }
 
     #[rustfmt::skip]
-    fn tag(&self) -> BerTag {
+    const fn tag(&self) -> Tag {
         match self {
-            BerObjectContent::EndOfContent         => BerTag::EndOfContent,
-            BerObjectContent::Boolean(_)           => BerTag::Boolean,
-            BerObjectContent::Integer(_)           => BerTag::Integer,
-            BerObjectContent::BitString(_,_)       => BerTag::BitString,
-            BerObjectContent::OctetString(_)       => BerTag::OctetString,
-            BerObjectContent::Null                 => BerTag::Null,
-            BerObjectContent::Enum(_)              => BerTag::Enumerated,
-            BerObjectContent::OID(_)               => BerTag::Oid,
-            BerObjectContent::NumericString(_)     => BerTag::NumericString,
-            BerObjectContent::VisibleString(_)     => BerTag::VisibleString,
-            BerObjectContent::PrintableString(_)   => BerTag::PrintableString,
-            BerObjectContent::IA5String(_)         => BerTag::Ia5String,
-            BerObjectContent::UTF8String(_)        => BerTag::Utf8String,
-            BerObjectContent::RelativeOID(_)       => BerTag::RelativeOid,
-            BerObjectContent::T61String(_)         => BerTag::T61String,
-            BerObjectContent::VideotexString(_)    => BerTag::VideotexString,
-            BerObjectContent::BmpString(_)         => BerTag::BmpString,
-            BerObjectContent::UniversalString(_)   => BerTag::UniversalString,
-            BerObjectContent::Sequence(_)          => BerTag::Sequence,
-            BerObjectContent::Set(_)               => BerTag::Set,
-            BerObjectContent::UTCTime(_)           => BerTag::UtcTime,
-            BerObjectContent::GeneralizedTime(_)   => BerTag::GeneralizedTime,
-            BerObjectContent::ObjectDescriptor(_)  => BerTag::ObjDescriptor,
-            BerObjectContent::GraphicString(_)     => BerTag::GraphicString,
-            BerObjectContent::GeneralString(_)     => BerTag::GeneralString,
-            BerObjectContent::Tagged(_,x,_) |
-            BerObjectContent::Unknown(x,_)         => *x,
+            BerObjectContent::EndOfContent         => Tag::EndOfContent,
+            BerObjectContent::Boolean(_)           => Tag::Boolean,
+            BerObjectContent::Integer(_)           => Tag::Integer,
+            BerObjectContent::BitString(_,_)       => Tag::BitString,
+            BerObjectContent::OctetString(_)       => Tag::OctetString,
+            BerObjectContent::Null                 => Tag::Null,
+            BerObjectContent::Enum(_)              => Tag::Enumerated,
+            BerObjectContent::OID(_)               => Tag::Oid,
+            BerObjectContent::NumericString(_)     => Tag::NumericString,
+            BerObjectContent::VisibleString(_)     => Tag::VisibleString,
+            BerObjectContent::PrintableString(_)   => Tag::PrintableString,
+            BerObjectContent::IA5String(_)         => Tag::Ia5String,
+            BerObjectContent::UTF8String(_)        => Tag::Utf8String,
+            BerObjectContent::RelativeOID(_)       => Tag::RelativeOid,
+            BerObjectContent::T61String(_)         => Tag::T61String,
+            BerObjectContent::VideotexString(_)    => Tag::VideotexString,
+            BerObjectContent::BmpString(_)         => Tag::BmpString,
+            BerObjectContent::UniversalString(_)   => Tag::UniversalString,
+            BerObjectContent::Sequence(_)          => Tag::Sequence,
+            BerObjectContent::Set(_)               => Tag::Set,
+            BerObjectContent::UTCTime(_)           => Tag::UtcTime,
+            BerObjectContent::GeneralizedTime(_)   => Tag::GeneralizedTime,
+            BerObjectContent::ObjectDescriptor(_)  => Tag::ObjectDescriptor,
+            BerObjectContent::GraphicString(_)     => Tag::GraphicString,
+            BerObjectContent::GeneralString(_)     => Tag::GeneralString,
+            BerObjectContent::Tagged(_,x,_) => *x,
+            BerObjectContent::Unknown(any) => any.tag(),
             BerObjectContent::Optional(Some(obj))  => obj.content.tag(),
-            BerObjectContent::Optional(None)       => BerTag(0x00), // XXX invalid !
+            BerObjectContent::Optional(None)       => Tag(0x00), // XXX invalid !
         }
     }
 }
@@ -914,12 +680,12 @@ impl<'a> BerObject<'a> {
     ///
     /// let (_, object) = parse_ber_integer(data).expect("parsing failed");
     /// # #[cfg(feature = "bigint")]
-    /// assert_eq!(object.as_bigint(), Some(65537.into()))
+    /// assert_eq!(object.as_bigint(), Ok(65537.into()))
     /// ```
-    pub fn as_bigint(&self) -> Option<BigInt> {
+    pub fn as_bigint(&self) -> Result<BigInt, BerError> {
         match self.content {
-            BerObjectContent::Integer(s) => Some(BigInt::from_signed_bytes_be(s)),
-            _ => None,
+            BerObjectContent::Integer(s) => Ok(BigInt::from_signed_bytes_be(s)),
+            _ => Err(BerError::BerValueError),
         }
     }
 
@@ -936,17 +702,17 @@ impl<'a> BerObject<'a> {
     ///
     /// let (_, object) = parse_ber_integer(data).expect("parsing failed");
     /// # #[cfg(feature = "bigint")]
-    /// assert_eq!(object.as_biguint(), Some(65537_u32.into()))
+    /// assert_eq!(object.as_biguint(), Ok(65537_u32.into()))
     /// ```
-    pub fn as_biguint(&self) -> Option<BigUint> {
+    pub fn as_biguint(&self) -> Result<BigUint, BerError> {
         match self.content {
             BerObjectContent::Integer(s) => {
                 if is_highest_bit_set(s) {
-                    return None;
+                    return Err(BerError::IntegerNegative);
                 }
-                Some(BigUint::from_bytes_be(s))
+                Ok(BigUint::from_bytes_be(s))
             }
-            _ => None,
+            _ => Err(BerError::BerValueError),
         }
     }
 }
@@ -1035,7 +801,7 @@ impl<'a> Index<usize> for BerObject<'a> {
         match (*self).content {
             BerObjectContent::Sequence(ref v) if idx < v.len() => &v[idx],
             BerObjectContent::Set(ref v) if idx < v.len() => &v[idx],
-            _ => panic!("Try to index BerObjectContent which is not structured"),
+            _ => panic!("Try to index BerObjectContent which is not constructed"),
         }
         // XXX the following
         // self.ref_iter().nth(idx).unwrap()
@@ -1063,8 +829,9 @@ impl<'a> BitStringObject<'a> {
     }
 
     /// Constructs a shared `&BitSlice` reference over the object data.
+    #[cfg(feature = "bitvec")]
     pub fn as_bitslice(&self) -> Option<&BitSlice<Msb0, u8>> {
-        BitSlice::<Msb0, _>::from_slice(self.data)
+        BitSlice::<Msb0, _>::from_slice(self.data).ok()
     }
 }
 
@@ -1134,16 +901,18 @@ mod tests {
         assert!(obj.is_set(17));
     }
 
+    #[cfg(feature = "bitvec")]
     #[test]
     fn test_der_bitslice() {
+        use std::string::String;
         let obj = BitStringObject {
             data: &[0x0f, 0x00, 0x40],
         };
         let slice = obj.as_bitslice().expect("as_bitslice");
-        assert_eq!(slice.get(0), Some(&false));
-        assert_eq!(slice.get(7), Some(&true));
-        assert_eq!(slice.get(9), Some(&false));
-        assert_eq!(slice.get(17), Some(&true));
+        assert_eq!(slice.get(0).as_deref(), Some(&false));
+        assert_eq!(slice.get(7).as_deref(), Some(&true));
+        assert_eq!(slice.get(9).as_deref(), Some(&false));
+        assert_eq!(slice.get(17).as_deref(), Some(&true));
         let s = slice.iter().fold(String::with_capacity(24), |mut acc, b| {
             acc += if *b { "1" } else { "0" };
             acc
@@ -1167,7 +936,7 @@ mod tests {
         let obj = BerObject::from_obj(BerObjectContent::Integer(b"\x01\x00\x01"));
         let expected = ::num_bigint::BigInt::from(0x10001);
 
-        assert_eq!(obj.as_bigint(), Some(expected));
+        assert_eq!(obj.as_bigint(), Ok(expected));
     }
 
     #[cfg(feature = "bigint")]
@@ -1176,6 +945,6 @@ mod tests {
         let obj = BerObject::from_obj(BerObjectContent::Integer(b"\x01\x00\x01"));
         let expected = ::num_bigint::BigUint::from(0x10001_u32);
 
-        assert_eq!(obj.as_biguint(), Some(expected));
+        assert_eq!(obj.as_biguint(), Ok(expected));
     }
 }
