@@ -1,5 +1,7 @@
+#![cfg(feature = "std")]
 use crate::ber::*;
 use crate::oid::Oid;
+use asn1_rs::{ASN1DateTime, Tag};
 use cookie_factory::bytes::be_u8;
 use cookie_factory::combinator::slice;
 use cookie_factory::gen_simple;
@@ -8,13 +10,11 @@ use cookie_factory::sequence::tuple;
 use cookie_factory::{GenError, SerializeFn};
 use std::io::Write;
 
-// we do not use .copied() for compatibility with 1.34
-#[allow(clippy::map_clone)]
-fn encode_length<'a, W: Write + 'a, Len: Into<BerSize>>(len: Len) -> impl SerializeFn<W> + 'a {
+fn encode_length<'a, W: Write + 'a, Len: Into<Length>>(len: Len) -> impl SerializeFn<W> + 'a {
     let l = len.into();
     move |out| {
         match l {
-            BerSize::Definite(sz) => {
+            Length::Definite(sz) => {
                 if sz <= 128 {
                     // definite, short form
                     be_u8(sz as u8)(out)
@@ -23,14 +23,14 @@ fn encode_length<'a, W: Write + 'a, Len: Into<BerSize>>(len: Len) -> impl Serial
                     let v: Vec<u8> = sz
                         .to_be_bytes()
                         .iter()
-                        .map(|&x| x)
+                        .cloned()
                         .skip_while(|&b| b == 0)
                         .collect();
                     let b0 = 0b1000_0000 | (v.len() as u8);
                     tuple((be_u8(b0 as u8), slice(v)))(out)
                 }
             }
-            BerSize::Indefinite => be_u8(0b1000_0000)(out),
+            Length::Indefinite => be_u8(0b1000_0000)(out),
         }
     }
 }
@@ -39,26 +39,31 @@ fn encode_length<'a, W: Write + 'a, Len: Into<BerSize>>(len: Len) -> impl Serial
 ///
 /// The `len` field must be correct
 #[cfg_attr(docsrs, doc(cfg(feature = "serialize")))]
-pub fn ber_encode_header<'a, 'b: 'a, W: Write + 'a>(
-    hdr: &'b BerObjectHeader,
-) -> impl SerializeFn<W> + 'a {
+pub fn ber_encode_header<'a, 'b: 'a, W: Write + 'a>(hdr: &'b Header) -> impl SerializeFn<W> + 'a {
     move |out| {
         // identifier octets (X.690 8.1.2)
-        let class_u8 = (hdr.class as u8) << 6;
-        let pc_u8 = (hdr.structured & 1) << 5;
-        if hdr.tag.0 >= 30 {
+        let class_u8 = (hdr.class() as u8) << 6;
+        let pc_u8 = (if hdr.constructed() { 1 } else { 0 }) << 5;
+        if hdr.tag().0 >= 30 {
             unimplemented!();
         }
-        let byte_0 = class_u8 | pc_u8 | (hdr.tag.0 as u8);
+        let byte_0 = class_u8 | pc_u8 | (hdr.tag().0 as u8);
         // length octets (X.690 8.1.3)
-        tuple((be_u8(byte_0), encode_length(hdr.len)))(out)
+        tuple((be_u8(byte_0), encode_length(hdr.length())))(out)
     }
 }
 
 fn ber_encode_oid<'a, W: Write + 'a>(oid: &'a Oid) -> impl SerializeFn<W> + 'a {
     move |out| {
         // check oid.relative attribute ? this should not be necessary
-        slice(oid.bytes())(out)
+        slice(oid.as_bytes())(out)
+    }
+}
+
+fn ber_encode_datetime<'a, W: Write + 'a>(time: &'a ASN1DateTime) -> impl SerializeFn<W> + 'a {
+    move |out| {
+        let s = format!("{}", time);
+        slice(s)(out)
     }
 }
 
@@ -73,8 +78,8 @@ fn ber_encode_sequence<'a, W: Write + Default + AsRef<[u8]> + 'a>(
 /// Note: `obj` should be the object to be encapsulated, not the `ContextSpecific` variant.
 #[cfg_attr(docsrs, doc(cfg(feature = "serialize")))]
 pub fn ber_encode_tagged_explicit<'a, W: Write + Default + AsRef<[u8]> + 'a>(
-    tag: BerTag,
-    class: BerClass,
+    tag: Tag,
+    class: Class,
     obj: &'a BerObject,
 ) -> impl SerializeFn<W> + 'a {
     move |out| {
@@ -82,7 +87,7 @@ pub fn ber_encode_tagged_explicit<'a, W: Write + Default + AsRef<[u8]> + 'a>(
         let v = gen_simple(ber_encode_object(obj), W::default())?;
         let len = v.as_ref().len();
         // encode the application header, using the tag
-        let hdr = BerObjectHeader::new(class, 1 /* X.690 8.14.2 */, tag, len);
+        let hdr = Header::new(class, true /* X.690 8.14.2 */, tag, len.into());
         let v_hdr = gen_simple(ber_encode_header(&hdr), W::default())?;
         tuple((slice(v_hdr), slice(v)))(out)
     }
@@ -93,23 +98,21 @@ pub fn ber_encode_tagged_explicit<'a, W: Write + Default + AsRef<[u8]> + 'a>(
 /// Note: `obj` should be the object to be encapsulated, not the `ContextSpecific` variant.
 #[cfg_attr(docsrs, doc(cfg(feature = "serialize")))]
 pub fn ber_encode_tagged_implicit<'a, W: Write + Default + AsRef<[u8]> + 'a>(
-    tag: BerTag,
-    class: BerClass,
+    tag: Tag,
+    class: Class,
     obj: &'a BerObject,
 ) -> impl SerializeFn<W> + 'a {
     move |out| {
         // encode inner object content
         let v = gen_simple(ber_encode_object_content(&obj.content), W::default())?;
-        // but replace the tag (keep structured attribute)
+        // but replace the tag (keep constructed attribute)
         let len = v.as_ref().len();
-        let hdr = BerObjectHeader::new(class, obj.header.structured, tag, len);
+        let hdr = Header::new(class, obj.header.constructed(), tag, len.into());
         let v_hdr = gen_simple(ber_encode_header(&hdr), W::default())?;
         tuple((slice(v_hdr), slice(v)))(out)
     }
 }
 
-// we do not use .copied() for compatibility with 1.34
-#[allow(clippy::map_clone)]
 fn ber_encode_object_content<'a, W: Write + Default + AsRef<[u8]> + 'a>(
     c: &'a BerObjectContent,
 ) -> impl SerializeFn<W> + 'a {
@@ -129,26 +132,27 @@ fn ber_encode_object_content<'a, W: Write + Default + AsRef<[u8]> + 'a>(
             let v: Vec<u8> = i
                 .to_be_bytes()
                 .iter()
-                .map(|&x| x)
+                .cloned()
                 .skip_while(|&b| b == 0)
                 .collect();
             slice(v)(out)
         }
         BerObjectContent::OID(oid) | BerObjectContent::RelativeOID(oid) => ber_encode_oid(oid)(out),
+        BerObjectContent::UTCTime(time) | BerObjectContent::GeneralizedTime(time) => {
+            ber_encode_datetime(time)(out)
+        }
         BerObjectContent::NumericString(s)
-        | BerObjectContent::UTCTime(s)
-        | BerObjectContent::GeneralizedTime(s)
+        | BerObjectContent::BmpString(s)
+        | BerObjectContent::GeneralString(s)
+        | BerObjectContent::ObjectDescriptor(s)
+        | BerObjectContent::GraphicString(s)
         | BerObjectContent::VisibleString(s)
         | BerObjectContent::PrintableString(s)
         | BerObjectContent::IA5String(s)
-        | BerObjectContent::UTF8String(s) => slice(s)(out),
-        BerObjectContent::T61String(s)
+        | BerObjectContent::T61String(s)
         | BerObjectContent::VideotexString(s)
-        | BerObjectContent::BmpString(s)
-        | BerObjectContent::UniversalString(s)
-        | BerObjectContent::ObjectDescriptor(s)
-        | BerObjectContent::GraphicString(s)
-        | BerObjectContent::GeneralString(s) => slice(s)(out),
+        | BerObjectContent::UTF8String(s) => slice(s)(out),
+        BerObjectContent::UniversalString(s) => slice(s)(out),
         BerObjectContent::Sequence(v) | BerObjectContent::Set(v) => ber_encode_sequence(v)(out),
         // best we can do is tagged-explicit, but we don't know
         BerObjectContent::Optional(inner) => {
@@ -163,7 +167,7 @@ fn ber_encode_object_content<'a, W: Write + Default + AsRef<[u8]> + 'a>(
             // XXX wrong, we should wrap it!
             ber_encode_object(inner)(out)
         }
-        BerObjectContent::Unknown(_tag, s) => slice(s)(out),
+        BerObjectContent::Unknown(any) => slice(any.data)(out),
     }
 }
 
@@ -175,7 +179,7 @@ fn ber_encode_object_content<'a, W: Write + Default + AsRef<[u8]> + 'a>(
 /// *This function is only available if the `serialize` feature is enabled.*
 #[cfg_attr(docsrs, doc(cfg(feature = "serialize")))]
 pub fn ber_encode_object_raw<'a, 'b: 'a, 'c: 'a, W: Write + Default + AsRef<[u8]> + 'a>(
-    hdr: &'b BerObjectHeader,
+    hdr: &'b Header,
     content: &'c BerObjectContent,
 ) -> impl SerializeFn<W> + 'a {
     tuple((ber_encode_header(hdr), ber_encode_object_content(content)))
@@ -197,7 +201,7 @@ pub fn ber_encode_object<'a, 'b: 'a, W: Write + Default + AsRef<[u8]> + 'a>(
         // XXX should we make an exception for tagged values here ?
         let v = gen_simple(ber_encode_object_content(&obj.content), W::default())?;
         let len = v.as_ref().len();
-        let hdr = obj.header.clone().with_len(len.into());
+        let hdr = obj.header.clone().with_length(len.into());
         let v_hdr = gen_simple(ber_encode_header(&hdr), W::default())?;
         tuple((slice(v_hdr), slice(v)))(out)
     }
@@ -346,12 +350,12 @@ mod test {
     #[test]
     fn test_encode_tagged_explicit() {
         fn local_parse(i: &[u8]) -> BerResult {
-            parse_ber_explicit_optional(i, BerTag(0), parse_ber_integer)
+            parse_ber_explicit_optional(i, Tag(0), parse_ber_integer)
         }
         let bytes = hex!("a0 03 02 01 02");
         let obj = BerObject::from_int_slice(b"\x02");
         let v = gen_simple(
-            ber_encode_tagged_explicit(BerTag(0), BerClass::ContextSpecific, &obj),
+            ber_encode_tagged_explicit(Tag(0), Class::ContextSpecific, &obj),
             Vec::new(),
         )
         .expect("could not encode");
@@ -361,7 +365,7 @@ mod test {
             .expect("tagged object not found")
             .expect("optional object empty");
         let (_class, tag, inner) = obj2.as_tagged().expect("not a tagged object");
-        assert_eq!(tag, BerTag(0));
+        assert_eq!(tag, Tag(0));
         assert_eq!(&obj, inner);
         assert_eq!(&v[..], bytes);
     }
@@ -370,22 +374,22 @@ mod test {
     fn test_encode_tagged_implicit() {
         fn der_read_integer_content<'a>(
             i: &'a [u8],
-            hdr: &BerObjectHeader,
+            hdr: &Header,
             depth: usize,
         ) -> BerResult<'a, BerObjectContent<'a>> {
-            ber_read_element_content_as(i, BerTag::Integer, hdr.len, false, depth)
+            ber_read_element_content_as(i, Tag::Integer, hdr.length(), false, depth)
         }
         fn local_parse(i: &[u8]) -> BerResult<BerObject> {
-            parse_ber_implicit(i, BerTag(3), der_read_integer_content)
+            parse_ber_implicit(i, Tag(3), der_read_integer_content)
         }
         let obj = BerObject::from_int_slice(b"\x02");
         let v = gen_simple(
-            ber_encode_tagged_implicit(BerTag(3), BerClass::ContextSpecific, &obj),
+            ber_encode_tagged_implicit(Tag(3), Class::ContextSpecific, &obj),
             Vec::new(),
         )
         .expect("could not encode");
         let (_, obj2) = local_parse(&v).expect("could not re-parse");
-        assert_eq!(obj2.header.tag, BerTag(3));
+        assert_eq!(obj2.header.tag(), Tag(3));
         assert_eq!(&obj.content, &obj2.content);
         let bytes = hex!("83 01 02");
         assert_eq!(&v[..], bytes);
@@ -393,11 +397,11 @@ mod test {
     #[test]
     fn test_encode_tagged_application() {
         fn local_parse(i: &[u8]) -> BerResult {
-            parse_ber_explicit_optional(i, BerTag(2), parse_ber_integer)
+            parse_ber_explicit_optional(i, Tag(2), parse_ber_integer)
         }
         let obj = BerObject::from_int_slice(b"\x02");
         let v = gen_simple(
-            ber_encode_tagged_explicit(BerTag(2), BerClass::Application, &obj),
+            ber_encode_tagged_explicit(Tag(2), Class::Application, &obj),
             Vec::new(),
         )
         .expect("could not encode");
@@ -407,7 +411,7 @@ mod test {
             .expect("tagged object not found")
             .expect("optional object empty");
         let (_class, tag, inner) = obj2.as_tagged().expect("not a tagged object");
-        assert_eq!(tag, BerTag(2));
+        assert_eq!(tag, Tag(2));
         assert_eq!(&obj, inner);
         let bytes = hex!("62 03 02 01 02");
         assert_eq!(&v[..], bytes);
