@@ -1,11 +1,13 @@
 use crate::ber::*;
 use crate::der::*;
+use crate::der_constraint_fail_if;
 use crate::error::*;
+use alloc::borrow::ToOwned;
+use asn1_rs::{Any, FromDer, Tag};
 use nom::bytes::streaming::take;
 use nom::number::streaming::be_u8;
 use nom::{Err, Needed};
 use rusticata_macros::custom_check;
-use std::convert::{Into, TryFrom};
 
 /// Parse DER object recursively
 ///
@@ -16,12 +18,12 @@ use std::convert::{Into, TryFrom};
 /// ### Example
 ///
 /// ```
-/// use der_parser::der::{parse_der, DerTag};
+/// use der_parser::der::{parse_der, Tag};
 ///
 /// let bytes = &[0x02, 0x03, 0x01, 0x00, 0x01];
 /// let (_, obj) = parse_der(bytes).expect("parsing failed");
 ///
-/// assert_eq!(obj.header.tag, DerTag::Integer);
+/// assert_eq!(obj.header.tag(), Tag::Integer);
 /// ```
 #[inline]
 pub fn parse_der(i: &[u8]) -> DerResult {
@@ -35,34 +37,22 @@ pub fn parse_der(i: &[u8]) -> DerResult {
 /// ### Example
 ///
 /// ```
-/// use der_parser::der::{parse_der_recursive, DerTag};
+/// use der_parser::der::{parse_der_recursive, Tag};
 ///
 /// let bytes = &[0x02, 0x03, 0x01, 0x00, 0x01];
 /// let (_, obj) = parse_der_recursive(bytes, 1).expect("parsing failed");
 ///
-/// assert_eq!(obj.header.tag, DerTag::Integer);
+/// assert_eq!(obj.header.tag(), Tag::Integer);
 /// ```
 pub fn parse_der_recursive(i: &[u8], max_depth: usize) -> DerResult {
     let (i, hdr) = der_read_element_header(i)?;
     // safety check: length cannot be more than 2^32 bytes
-    if let BerSize::Definite(l) = hdr.len {
+    if let Length::Definite(l) = hdr.length() {
         custom_check!(i, l > MAX_OBJECT_SIZE, BerError::InvalidLength)?;
     }
     der_read_element_content_recursive(i, hdr, max_depth)
 }
 
-#[doc(hidden)]
-#[macro_export]
-macro_rules! der_constraint_fail_if(
-    ($slice:expr, $cond:expr) => (
-        {
-            if $cond {
-                return Err(::nom::Err::Error(BerError::DerConstraintFailed));
-            }
-        }
-    );
-);
-
 /// Parse a DER object, expecting a value with specified tag
 ///
 /// The object is parsed recursively, with a maximum depth of `MAX_RECURSION`.
@@ -70,28 +60,31 @@ macro_rules! der_constraint_fail_if(
 /// ### Example
 ///
 /// ```
-/// use der_parser::der::{parse_der_with_tag, DerTag};
+/// use der_parser::der::{parse_der_with_tag, Tag};
 ///
 /// let bytes = &[0x02, 0x03, 0x01, 0x00, 0x01];
-/// let (_, obj) = parse_der_with_tag(bytes, DerTag::Integer).expect("parsing failed");
+/// let (_, obj) = parse_der_with_tag(bytes, Tag::Integer).expect("parsing failed");
 ///
-/// assert_eq!(obj.header.tag, DerTag::Integer);
+/// assert_eq!(obj.header.tag(), Tag::Integer);
 /// ```
-pub fn parse_der_with_tag<Tag: Into<DerTag>>(i: &[u8], tag: Tag) -> DerResult {
+pub fn parse_der_with_tag<T: Into<Tag>>(i: &[u8], tag: T) -> DerResult {
     let tag = tag.into();
     let (i, hdr) = der_read_element_header(i)?;
-    if hdr.tag != tag {
-        return Err(nom::Err::Error(BerError::InvalidTag));
-    }
-    let (i, content) =
-        der_read_element_content_as(i, hdr.tag, hdr.len, hdr.is_constructed(), MAX_RECURSION)?;
+    hdr.assert_tag(tag)?;
+    let (i, content) = der_read_element_content_as(
+        i,
+        hdr.tag(),
+        hdr.length(),
+        hdr.is_constructed(),
+        MAX_RECURSION,
+    )?;
     Ok((i, DerObject::from_header_and_content(hdr, content)))
 }
 
 /// Read end of content marker
 #[inline]
 pub fn parse_der_endofcontent(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::EndOfContent)
+    parse_der_with_tag(i, Tag::EndOfContent)
 }
 
 /// Read a boolean value
@@ -103,7 +96,7 @@ pub fn parse_der_endofcontent(i: &[u8]) -> DerResult {
 /// If the boolean value is TRUE, the octet shall be one byte, and have all bits set to one (0xff).
 #[inline]
 pub fn parse_der_bool(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::Boolean)
+    parse_der_with_tag(i, Tag::Boolean)
 }
 
 /// Read an integer value
@@ -132,7 +125,7 @@ pub fn parse_der_bool(i: &[u8]) -> DerResult {
 /// ```
 #[inline]
 pub fn parse_der_integer(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::Integer)
+    parse_der_with_tag(i, Tag::Integer)
 }
 
 /// Read an bitstring value
@@ -153,150 +146,150 @@ pub fn parse_der_integer(i: &[u8]) -> DerResult {
 /// objects.
 #[inline]
 pub fn parse_der_bitstring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::BitString)
+    parse_der_with_tag(i, Tag::BitString)
 }
 
 /// Read an octetstring value
 #[inline]
 pub fn parse_der_octetstring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::OctetString)
+    parse_der_with_tag(i, Tag::OctetString)
 }
 
 /// Read a null value
 #[inline]
 pub fn parse_der_null(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::Null)
+    parse_der_with_tag(i, Tag::Null)
 }
 
 /// Read an object identifier value
 #[inline]
 pub fn parse_der_oid(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::Oid)
+    parse_der_with_tag(i, Tag::Oid)
 }
 
 /// Read an enumerated value
 #[inline]
 pub fn parse_der_enum(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::Enumerated)
+    parse_der_with_tag(i, Tag::Enumerated)
 }
 
 /// Read a UTF-8 string value. The encoding is checked.
 #[inline]
 pub fn parse_der_utf8string(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::Utf8String)
+    parse_der_with_tag(i, Tag::Utf8String)
 }
 
 /// Read a relative object identifier value
 #[inline]
 pub fn parse_der_relative_oid(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::RelativeOid)
+    parse_der_with_tag(i, Tag::RelativeOid)
 }
 
 /// Parse a sequence of DER elements
 ///
 /// Read a sequence of DER objects, without any constraint on the types.
-/// Sequence is parsed recursively, so if structured elements are found, they are parsed using the
+/// Sequence is parsed recursively, so if constructed elements are found, they are parsed using the
 /// same function.
 ///
 /// To read a specific sequence of objects (giving the expected types), use the
 /// [`parse_der_sequence_defined`](macro.parse_der_sequence_defined.html) macro.
 #[inline]
 pub fn parse_der_sequence(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::Sequence)
+    parse_der_with_tag(i, Tag::Sequence)
 }
 
 /// Parse a set of DER elements
 ///
 /// Read a set of DER objects, without any constraint on the types.
-/// Set is parsed recursively, so if structured elements are found, they are parsed using the
+/// Set is parsed recursively, so if constructed elements are found, they are parsed using the
 /// same function.
 ///
 /// To read a specific set of objects (giving the expected types), use the
 /// [`parse_der_set_defined`](macro.parse_der_set_defined.html) macro.
 #[inline]
 pub fn parse_der_set(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::Set)
+    parse_der_with_tag(i, Tag::Set)
 }
 
 /// Read a numeric string value. The content is verified to
 /// contain only digits and spaces.
 #[inline]
 pub fn parse_der_numericstring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::NumericString)
+    parse_der_with_tag(i, Tag::NumericString)
 }
 
 /// Read a printable string value. The content is verified to
 /// contain only the allowed characters.
 #[inline]
 pub fn visiblestring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::VisibleString)
+    parse_der_with_tag(i, Tag::VisibleString)
 }
 
 /// Read a printable string value. The content is verified to
 /// contain only the allowed characters.
 #[inline]
 pub fn parse_der_printablestring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::PrintableString)
+    parse_der_with_tag(i, Tag::PrintableString)
 }
 
 /// Read a T61 string value
 #[inline]
 pub fn parse_der_t61string(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::T61String)
+    parse_der_with_tag(i, Tag::T61String)
 }
 
 /// Read a Videotex string value
 #[inline]
 pub fn parse_der_videotexstring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::VideotexString)
+    parse_der_with_tag(i, Tag::VideotexString)
 }
 
 /// Read an IA5 string value. The content is verified to be ASCII.
 #[inline]
 pub fn parse_der_ia5string(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::Ia5String)
+    parse_der_with_tag(i, Tag::Ia5String)
 }
 
 /// Read an UTC time value
 #[inline]
 pub fn parse_der_utctime(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::UtcTime)
+    parse_der_with_tag(i, Tag::UtcTime)
 }
 
 /// Read a Generalized time value
 #[inline]
 pub fn parse_der_generalizedtime(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::GeneralizedTime)
+    parse_der_with_tag(i, Tag::GeneralizedTime)
 }
 
 /// Read a ObjectDescriptor value
 #[inline]
 pub fn parse_der_objectdescriptor(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::ObjDescriptor)
+    parse_der_with_tag(i, Tag::ObjectDescriptor)
 }
 
 /// Read a GraphicString value
 #[inline]
 pub fn parse_der_graphicstring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::GraphicString)
+    parse_der_with_tag(i, Tag::GraphicString)
 }
 
 /// Read a GeneralString value
 #[inline]
 pub fn parse_der_generalstring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::GeneralString)
+    parse_der_with_tag(i, Tag::GeneralString)
 }
 
 /// Read a BmpString value
 #[inline]
 pub fn parse_der_bmpstring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::BmpString)
+    parse_der_with_tag(i, Tag::BmpString)
 }
 
 /// Read a UniversalString value
 #[inline]
 pub fn parse_der_universalstring(i: &[u8]) -> DerResult {
-    parse_der_with_tag(i, DerTag::UniversalString)
+    parse_der_with_tag(i, Tag::UniversalString)
 }
 
 /// Parse an optional tagged object, applying function to get content
@@ -310,29 +303,13 @@ pub fn parse_der_universalstring(i: &[u8]) -> DerResult {
 /// This function will never fail: if parsing content failed, the BER value `Optional(None)` is
 /// returned.
 #[inline]
-pub fn parse_der_explicit_optional<F>(i: &[u8], tag: DerTag, f: F) -> DerResult
+pub fn parse_der_explicit_optional<F>(i: &[u8], tag: Tag, f: F) -> DerResult
 where
     F: Fn(&[u8]) -> DerResult,
 {
     parse_ber_explicit_optional(i, tag, f)
 }
 
-/// Parse an optional tagged object, applying function to get content
-///
-/// This function is deprecated, use
-/// [parse_der_explicit_optional](fn.parse_der_explicit_optional.html) instead.
-#[deprecated(
-    since = "4.1.0",
-    note = "Please use `parse_der_explicit_optional` instead"
-)]
-#[inline]
-pub fn parse_der_explicit<F>(i: &[u8], tag: DerTag, f: F) -> BerResult
-where
-    F: Fn(&[u8]) -> BerResult,
-{
-    parse_der_explicit_optional(i, tag, f)
-}
-
 /// Parse an implicit tagged object, applying function to read content
 ///
 /// Note: unlike explicit tagged functions, the callback must be a *content* parsing function,
@@ -358,7 +335,7 @@ where
 ///     parse_der_implicit(
 ///         i,
 ///         3,
-///         parse_der_content(DerTag::Integer),
+///         parse_der_content(Tag::Integer),
 ///     )
 /// }
 ///
@@ -373,10 +350,10 @@ where
 /// # }
 /// ```
 #[inline]
-pub fn parse_der_implicit<'a, Tag, F>(i: &'a [u8], tag: Tag, f: F) -> DerResult<'a>
+pub fn parse_der_implicit<'a, T, F>(i: &'a [u8], tag: T, f: F) -> DerResult<'a>
 where
-    F: Fn(&'a [u8], &'_ DerObjectHeader, usize) -> BerResult<'a, DerObjectContent<'a>>,
-    Tag: Into<DerTag>,
+    F: Fn(&'a [u8], &'_ Header, usize) -> BerResult<'a, DerObjectContent<'a>>,
+    T: Into<Tag>,
 {
     parse_ber_implicit(i, tag, f)
 }
@@ -387,9 +364,7 @@ where
 /// integer type.
 #[inline]
 pub fn parse_der_i32(i: &[u8]) -> BerResult<i32> {
-    let (rem, der) = parse_der_integer(i)?;
-    let int = der.as_i32().map_err(nom::Err::Error)?;
-    Ok((rem, int))
+    <i32>::from_der(i)
 }
 
 /// Parse DER object and try to decode it as a 64-bits signed integer
@@ -398,9 +373,7 @@ pub fn parse_der_i32(i: &[u8]) -> BerResult<i32> {
 /// integer type.
 #[inline]
 pub fn parse_der_i64(i: &[u8]) -> BerResult<i64> {
-    let (rem, der) = parse_der_integer(i)?;
-    let int = der.as_i64().map_err(nom::Err::Error)?;
-    Ok((rem, int))
+    <i64>::from_der(i)
 }
 
 /// Parse DER object and try to decode it as a 32-bits unsigned integer
@@ -408,9 +381,7 @@ pub fn parse_der_i64(i: &[u8]) -> BerResult<i64> {
 /// Return `IntegerTooLarge` if object is an integer, but can not be represented in the target
 /// integer type.
 pub fn parse_der_u32(i: &[u8]) -> BerResult<u32> {
-    let (rem, der) = parse_der_integer(i)?;
-    let int = der.as_u32().map_err(nom::Err::Error)?;
-    Ok((rem, int))
+    <u32>::from_der(i)
 }
 
 /// Parse DER object and try to decode it as a 64-bits unsigned integer
@@ -418,19 +389,15 @@ pub fn parse_der_u32(i: &[u8]) -> BerResult<u32> {
 /// Return `IntegerTooLarge` if object is an integer, but can not be represented in the target
 /// integer type.
 pub fn parse_der_u64(i: &[u8]) -> BerResult<u64> {
-    let (rem, der) = parse_der_integer(i)?;
-    let int = der.as_u64().map_err(nom::Err::Error)?;
-    Ok((rem, int))
+    <u64>::from_der(i)
 }
 
 /// Parse DER object and get content as slice
 #[inline]
-pub fn parse_der_slice<Tag: Into<DerTag>>(i: &[u8], tag: Tag) -> BerResult<&[u8]> {
+pub fn parse_der_slice<T: Into<Tag>>(i: &[u8], tag: T) -> BerResult<&[u8]> {
     let tag = tag.into();
     parse_der_container(move |content, hdr| {
-        if hdr.tag != tag {
-            return Err(Err::Error(BerError::InvalidTag));
-        }
+        hdr.assert_tag(tag)?;
         Ok((&b""[..], content))
     })(i)
 }
@@ -455,16 +422,16 @@ pub fn parse_der_slice<Tag: Into<DerTag>>(i: &[u8], tag: Tag) -> BerResult<&[u8]
 /// #
 /// # let bytes = &[0x02, 0x03, 0x01, 0x00, 0x01];
 /// let (i, header) = der_read_element_header(bytes).expect("parsing failed");
-/// let (rem, content) = parse_der_content(header.tag)(i, &header, MAX_RECURSION)
+/// let (rem, content) = parse_der_content(header.tag())(i, &header, MAX_RECURSION)
 ///     .expect("parsing failed");
 /// #
-/// # assert_eq!(header.tag, DerTag::Integer);
+/// # assert_eq!(header.tag(), Tag::Integer);
 /// ```
 pub fn parse_der_content<'a>(
-    tag: DerTag,
-) -> impl Fn(&'a [u8], &'_ DerObjectHeader, usize) -> BerResult<'a, DerObjectContent<'a>> {
-    move |i: &[u8], hdr: &DerObjectHeader, max_recursion: usize| {
-        der_read_element_content_as(i, tag, hdr.len, hdr.is_constructed(), max_recursion)
+    tag: Tag,
+) -> impl Fn(&'a [u8], &'_ Header, usize) -> BerResult<'a, DerObjectContent<'a>> {
+    move |i: &[u8], hdr: &Header, max_recursion: usize| {
+        der_read_element_content_as(i, tag, hdr.length(), hdr.is_constructed(), max_recursion)
     }
 }
 
@@ -488,15 +455,15 @@ pub fn parse_der_content<'a>(
 /// #
 /// # let bytes = &[0x02, 0x03, 0x01, 0x00, 0x01];
 /// let (i, header) = der_read_element_header(bytes).expect("parsing failed");
-/// # assert_eq!(header.tag, DerTag::Integer);
-/// let (rem, content) = parse_der_content2(header.tag)(i, header, MAX_RECURSION)
+/// # assert_eq!(header.tag(), Tag::Integer);
+/// let (rem, content) = parse_der_content2(header.tag())(i, header, MAX_RECURSION)
 ///     .expect("parsing failed");
 /// ```
 pub fn parse_der_content2<'a>(
-    tag: DerTag,
-) -> impl Fn(&'a [u8], DerObjectHeader<'a>, usize) -> BerResult<'a, DerObjectContent<'a>> {
-    move |i: &[u8], hdr: DerObjectHeader, max_recursion: usize| {
-        der_read_element_content_as(i, tag, hdr.len, hdr.is_constructed(), max_recursion)
+    tag: Tag,
+) -> impl Fn(&'a [u8], Header<'a>, usize) -> BerResult<'a, DerObjectContent<'a>> {
+    move |i: &[u8], hdr: Header, max_recursion: usize| {
+        der_read_element_content_as(i, tag, hdr.length(), hdr.is_constructed(), max_recursion)
     }
 }
 
@@ -507,87 +474,101 @@ pub fn parse_der_content2<'a>(
 /// Content type is *not* checked, caller is responsible of providing the correct tag
 pub fn der_read_element_content_as(
     i: &[u8],
-    tag: DerTag,
-    len: BerSize,
+    tag: Tag,
+    length: Length,
     constructed: bool,
     max_depth: usize,
 ) -> BerResult<DerObjectContent> {
     // Indefinite lengths are not allowed in DER (X.690 section 10.1)
-    let l = len.primitive()?;
+    let l = length.definite().map_err(BerError::from)?;
     if i.len() < l {
         return Err(Err::Incomplete(Needed::new(l)));
     }
     match tag {
-        DerTag::Boolean => {
+        Tag::Boolean => {
             custom_check!(i, l != 1, BerError::InvalidLength)?;
-            der_constraint_fail_if!(i, i[0] != 0 && i[0] != 0xff);
+            der_constraint_fail_if!(i, i[0] != 0 && i[0] != 0xff, DerConstraint::InvalidBoolean);
         }
-        DerTag::BitString => {
-            der_constraint_fail_if!(i, constructed);
+        Tag::BitString => {
+            der_constraint_fail_if!(i, constructed, DerConstraint::Constructed);
             // exception: read and verify padding bits
             return der_read_content_bitstring(i, l);
         }
-        DerTag::Integer => {
+        Tag::Integer => {
             // verify leading zeros
             match i[..l] {
-                [] => return Err(nom::Err::Error(BerError::DerConstraintFailed)),
-                [0, 0, ..] => return Err(nom::Err::Error(BerError::DerConstraintFailed)),
+                [] => {
+                    return Err(nom::Err::Error(BerError::DerConstraintFailed(
+                        DerConstraint::IntegerEmpty,
+                    )))
+                }
+                [0, 0, ..] => {
+                    return Err(nom::Err::Error(BerError::DerConstraintFailed(
+                        DerConstraint::IntegerLeadingZeroes,
+                    )))
+                }
                 [0, byte, ..] if byte < 0x80 => {
-                    return Err(nom::Err::Error(BerError::DerConstraintFailed));
+                    return Err(nom::Err::Error(BerError::DerConstraintFailed(
+                        DerConstraint::IntegerLeadingZeroes,
+                    )));
                 }
                 _ => (),
             }
         }
-        DerTag::NumericString
-        | DerTag::VisibleString
-        | DerTag::PrintableString
-        | DerTag::Ia5String
-        | DerTag::Utf8String
-        | DerTag::T61String
-        | DerTag::VideotexString
-        | DerTag::BmpString
-        | DerTag::UniversalString
-        | DerTag::ObjDescriptor
-        | DerTag::GraphicString
-        | DerTag::GeneralString => {
-            der_constraint_fail_if!(i, constructed);
+        Tag::NumericString
+        | Tag::VisibleString
+        | Tag::PrintableString
+        | Tag::Ia5String
+        | Tag::Utf8String
+        | Tag::T61String
+        | Tag::VideotexString
+        | Tag::BmpString
+        | Tag::UniversalString
+        | Tag::ObjectDescriptor
+        | Tag::GraphicString
+        | Tag::GeneralString => {
+            der_constraint_fail_if!(i, constructed, DerConstraint::Constructed);
         }
-        DerTag::UtcTime | DerTag::GeneralizedTime => {
+        Tag::UtcTime | Tag::GeneralizedTime => {
             if l == 0 || i.get(l - 1).cloned() != Some(b'Z') {
-                return Err(Err::Error(BerError::DerConstraintFailed));
+                return Err(Err::Error(BerError::DerConstraintFailed(
+                    DerConstraint::MissingTimeZone,
+                )));
             }
         }
         _ => (),
     }
-    ber_read_element_content_as(i, tag, len, constructed, max_depth)
+    ber_read_element_content_as(i, tag, length, constructed, max_depth)
 }
 
 /// Parse DER object content recursively
 ///
 /// *Note: an error is raised if recursion depth exceeds `MAX_RECURSION`.
-pub fn der_read_element_content<'a>(i: &'a [u8], hdr: DerObjectHeader<'a>) -> DerResult<'a> {
+pub fn der_read_element_content<'a>(i: &'a [u8], hdr: Header<'a>) -> DerResult<'a> {
     der_read_element_content_recursive(i, hdr, MAX_RECURSION)
 }
 
 fn der_read_element_content_recursive<'a>(
     i: &'a [u8],
-    hdr: DerObjectHeader<'a>,
+    hdr: Header<'a>,
     max_depth: usize,
 ) -> DerResult<'a> {
-    match hdr.class {
-        DerClass::Universal | DerClass::Private => (),
+    match hdr.class() {
+        Class::Universal => (),
         _ => {
-            let (i, content) = ber_get_object_content(i, &hdr, max_depth)?;
-            let content = DerObjectContent::Unknown(hdr.tag, content);
+            let (i, data) = ber_get_object_content(i, &hdr, max_depth)?;
+            let any = Any::new(hdr.clone(), data);
+            let content = DerObjectContent::Unknown(any);
             let obj = DerObject::from_header_and_content(hdr, content);
             return Ok((i, obj));
         }
     }
-    match der_read_element_content_as(i, hdr.tag, hdr.len, hdr.is_constructed(), max_depth) {
+    match der_read_element_content_as(i, hdr.tag(), hdr.length(), hdr.is_constructed(), max_depth) {
         Ok((rem, content)) => Ok((rem, DerObject::from_header_and_content(hdr, content))),
-        Err(Err::Error(BerError::UnknownTag)) => {
-            let (rem, content) = ber_get_object_content(i, &hdr, max_depth)?;
-            let content = DerObjectContent::Unknown(hdr.tag, content);
+        Err(Err::Error(BerError::UnknownTag(_))) => {
+            let (rem, data) = ber_get_object_content(i, &hdr, max_depth)?;
+            let any = Any::new(hdr.clone(), data);
+            let content = DerObjectContent::Unknown(any);
             let obj = DerObject::from_header_and_content(hdr, content);
             Ok((rem, obj))
         }
@@ -598,7 +579,10 @@ fn der_read_element_content_recursive<'a>(
 fn der_read_content_bitstring(i: &[u8], len: usize) -> BerResult<DerObjectContent> {
     let (i, ignored_bits) = be_u8(i)?;
     if ignored_bits > 7 {
-        return Err(Err::Error(BerError::DerConstraintFailed));
+        return Err(Err::Error(BerError::invalid_value(
+            Tag::BitString,
+            "More than 7 unused bits".to_owned(),
+        )));
     }
     if len == 0 {
         return Err(Err::Error(BerError::InvalidLength));
@@ -607,7 +591,7 @@ fn der_read_content_bitstring(i: &[u8], len: usize) -> BerResult<DerObjectConten
     if len > 1 {
         let mut last_byte = data[len - 2];
         for _ in 0..ignored_bits as usize {
-            der_constraint_fail_if!(i, last_byte & 1 != 0);
+            der_constraint_fail_if!(i, last_byte & 1 != 0, DerConstraint::UnusedBitsNotZero);
             last_byte >>= 1;
         }
     }
@@ -636,44 +620,7 @@ fn der_read_content_bitstring(i: &[u8], len: usize) -> BerResult<DerObjectConten
 }
 
 /// Read an object header (DER)
-pub fn der_read_element_header(i: &[u8]) -> BerResult<DerObjectHeader> {
-    let (i1, el) = parse_identifier(i)?;
-    let class = match DerClass::try_from(el.0) {
-        Ok(c) => c,
-        Err(_) => unreachable!(), // Cannot fail, we have read exactly 2 bits
-    };
-    let (i2, len) = parse_ber_length_byte(i1)?;
-    let (i3, len) = match (len.0, len.1) {
-        (0, l1) => {
-            // Short form: MSB is 0, the rest encodes the length (which can be 0) (8.1.3.4)
-            (i2, BerSize::Definite(usize::from(l1)))
-        }
-        (_, 0) => {
-            // Indefinite form is not allowed in DER (10.1)
-            return Err(::nom::Err::Error(BerError::DerConstraintFailed));
-        }
-        (_, l1) => {
-            // if len is 0xff -> error (8.1.3.5)
-            if l1 == 0b0111_1111 {
-                return Err(::nom::Err::Error(BerError::InvalidTag));
-            }
-            // DER(9.1) if len is 0 (indefinite form), obj must be constructed
-            der_constraint_fail_if!(&i[1..], len.1 == 0 && el.1 != 1);
-            let (i3, llen) = take(l1)(i2)?;
-            match bytes_to_u64(llen) {
-                Ok(l) => {
-                    // DER: should have been encoded in short form (< 127)
-                    der_constraint_fail_if!(i, l < 127);
-                    let l =
-                        usize::try_from(l).or(Err(::nom::Err::Error(BerError::InvalidLength)))?;
-                    (i3, BerSize::Definite(l))
-                }
-                Err(_) => {
-                    return Err(::nom::Err::Error(BerError::InvalidTag));
-                }
-            }
-        }
-    };
-    let hdr = DerObjectHeader::new(class, el.1, BerTag(el.2), len).with_raw_tag(Some(el.3));
-    Ok((i3, hdr))
+#[inline]
+pub fn der_read_element_header(i: &[u8]) -> BerResult<Header> {
+    Header::from_der(i)
 }
